@@ -10,12 +10,12 @@ from django import forms
 from django.http.response import HttpResponse, HttpResponseRedirect
 from django.utils.html import format_html
 
-import requests
-
 from payments import PaymentStatus, RedirectNeeded
 from payments.core import BasicProvider, get_base_url
 from payments.forms import PaymentForm
 
+import requests
+
 
 sig_sorted_key_list = [
     "currency-code",
@@ -43,30 +43,63 @@ CURRENCY_SUB_UNIT = {
 CENTS = Decimal('0.01')
 
 
+def add_extra_data(payment, new_extra_data):
+    if payment.extra_data:
+        old_extra_data = json.loads(payment.extra_data)
+    else:
+        old_extra_data = {}
+    extra_data = {**old_extra_data, **new_extra_data}
+    payment.extra_data = json.dumps(extra_data, indent=2)
+    payment.save()
+
+
+def add_new_status(payment, new_status):
+    if payment.extra_data:
+        old_extra_data = json.loads(payment.extra_data)
+    else:
+        old_extra_data = {}
+    if 'statuses' not in old_extra_data:
+        old_extra_data['statuses'] = []
+    old_extra_data['statuses'].append(new_status)
+    payment.extra_data = json.dumps(old_extra_data, indent=2)
+    payment.save()
+
+
 def quantize_price(price, currency):
     price = price * CURRENCY_SUB_UNIT[currency]
     return int(price.quantize(CENTS, rounding=ROUND_HALF_UP))
 
 
-# A bit hacky method, how to get PayU javascript into the form
-class ScriptField(forms.HiddenInput):
-    def __init__(self, script_params, payu_base_url, payment, *args, **kwargs):
-        self.script_params = script_params
-        self.payu_base_url = payu_base_url
-        self.payment = payment
-        return super(ScriptField, self).__init__(*args, **kwargs)
+def dequantize_price(price, currency):
+    price = Decimal(price) / CURRENCY_SUB_UNIT[currency]
+    return price
 
-    def render(self, name, value, script_params={}, attrs=None, renderer=None):
-        super(ScriptField, self).render(name, value, attrs)
+
+# A bit hacky method, how to get html output instead of form (for PayU express form and error form)
+class HtmlOutputField(forms.HiddenInput):
+    def __init__(self, *args, html='', **kwargs):
+        self.html = html
+        return super(HtmlOutputField, self).__init__(*args, **kwargs)
+
+    def render(self, *args, **kwargs):
+        return self.html
+
+
+class WidgetPaymentForm(PaymentForm):
+    hide_submit_button = True  # For easy use in templates
+    script = forms.CharField(label="Script")
+
+    def __init__(self, payu_base_url, script_params={}, *args, **kwargs):
+        ret = super(WidgetPaymentForm, self).__init__(*args, **kwargs)
         inline_code = format_html(
             "<script "
-            f"src='{self.payu_base_url}front/widget/js/payu-bootstrap.js' "
+            f"src='{payu_base_url}front/widget/js/payu-bootstrap.js' "
             "pay-button='#pay-button' {} >"
             "</script>",
-            " ".join('%s=%s' % (k, v) for k, v in self.script_params.items()),
+            " ".join('%s=%s' % (k, v) for k, v in script_params.items()),
         )
 
-        return inline_code + """
+        form_html = inline_code + """
             <script>
                 function cardSuccess($data) {
                     console.log('callback');
@@ -94,19 +127,7 @@ class ScriptField(forms.HiddenInput):
                 self.payment.get_success_url(),
             ),
         )
-
-
-class WidgetPaymentForm(PaymentForm):
-    hide_submit_button = True  # For easy use in templates
-    script = forms.CharField(label="Script")
-
-    def __init__(self, payu_base_url, script_params={}, *args, **kwargs):
-        ret = super(WidgetPaymentForm, self).__init__(*args, **kwargs)
-        self.fields['script'].widget = ScriptField(
-            payu_base_url=payu_base_url,
-            script_params=script_params,
-            payment=self.payment,
-        )
+        self.fields['script'].widget = HtmlOutputField(html=form_html)
         return ret
 
 
@@ -119,6 +140,16 @@ class RenewPaymentForm(PaymentForm):
         return ret
 
 
+class PaymentErrorForm(forms.Form):
+    script = forms.CharField(
+        widget=HtmlOutputField(
+            html='<br/><strong>This payment is already being processed.<br/></strong>',
+        ),
+    )
+    hide_submit_button = True
+    error_form = True
+
+
 class PayuApiError(Exception):
     pass
 
@@ -177,8 +208,12 @@ class PayuProvider(BasicProvider):
         cvv_url = None
         if payment.extra_data:
             extra_data = json.loads(payment.extra_data)
-            if 'cvv_url' in extra_data:
-                cvv_url = extra_data['cvv_url']
+            if '3ds_url' in extra_data:
+                raise RedirectNeeded(extra_data['3ds_url'])
+            cvv_url = extra_data.get('cvv_url', None)
+
+        if payment.status != PaymentStatus.WAITING:
+            return PaymentErrorForm()
 
         renew_token = payment.get_renew_token()
         if renew_token and self.recurring_payments and not cvv_url:
@@ -198,7 +233,7 @@ class PayuProvider(BasicProvider):
             payu_data.update({
                 'cvv-url': cvv_url,
                 'cvv-success-callback': 'cvvSuccess',
-                'widget-type': 'cvv'
+                'widget-type': 'cvv',
             })
         else:
             payu_data.update({
@@ -209,7 +244,14 @@ class PayuProvider(BasicProvider):
             if self.recurring_payments:
                 payu_data["recurring-payment"] = "true"
         payu_data['sig'] = self.get_sig(payu_data)
-        return WidgetPaymentForm(payu_base_url=self.payu_base_url, data=data, script_params=payu_data, provider=self, payment=payment)
+
+        return WidgetPaymentForm(
+            payu_base_url=self.payu_base_url,
+            data=data,
+            script_params=payu_data,
+            provider=self,
+            payment=payment,
+        )
 
     def get_processor(self, payment):
         order = payment.get_purchased_items()
@@ -221,6 +263,7 @@ class PayuProvider(BasicProvider):
             description=payment.description,
             customer_ip=payment.customer_ip_address,
             total=payment.total,
+            tax=payment.tax,
         )
         processor.set_buyer_data(
             first_name=payment.get_user_first_name(),
@@ -236,6 +279,7 @@ class PayuProvider(BasicProvider):
         processor = self.get_processor(payment)
         if self.card_on_file:
             processor.cardOnFile = 'FIRST' if recurring == 'FIRST' else 'STANDARD_CARDHOLDER'
+            # TODO: or STANDARD_MERCHANT
         elif self.recurring_payments:
             processor.recurring = recurring
         if self.express_payments:
@@ -250,9 +294,8 @@ class PayuProvider(BasicProvider):
         return HttpResponse(data, status=200)
 
     def post_request(self, url, *args, **kwargs):
-        if 'headers' not in kwargs:
-            kwargs['headers'] = self.get_token_headers()
         for i in range(1, self.retry_count):
+            kwargs['headers'] = self.get_token_headers()
             response = requests.post(url, *args, **kwargs)
             response_dict = json.loads(response.text)
             if (
@@ -263,10 +306,10 @@ class PayuProvider(BasicProvider):
                 try:
                     self.token = self.get_access_token(self.pos_id, self.client_secret, grant_type=self.grant_type)
                 except PayuApiError as e:
-                    pass
+                    raise PayuApiError(f"Unable to regain authorization token {e}")
             else:
                 return response_dict
-        raise PayuApiError(f"Unable to regain authorization token {e}")
+        raise PayuApiError("Unable to regain authorization token")
 
     def get_access_token(
             self,
@@ -297,7 +340,11 @@ class PayuProvider(BasicProvider):
             'Content-Type': 'application/x-www-form-urlencoded',
         }
         response = requests.post(payu_auth_url, data=data, headers=headers)
-        response_dict = json.loads(response.text)
+
+        try:
+            response_dict = json.loads(response.text)
+        except json.JSONDecodeError:
+            raise PayuApiError(response.text)
 
         try:
             return response_dict['access_token']
@@ -315,7 +362,6 @@ class PayuProvider(BasicProvider):
 
         payu_delete_token_url = urljoin(self.payu_token_url, card_token)
         response = requests.delete(payu_delete_token_url, headers=self.get_token_headers())
-        self.token = None
 
         return (response.status_code == 204)
 
@@ -325,27 +371,29 @@ class PayuProvider(BasicProvider):
 
         return redirectUrl  url where the user should go next
         """
-
         payment = payment
         payment_processor.pos_id = self.pos_id
+        json_data = payment_processor.as_json()
         response_dict = self.post_request(
             self.payu_api_order_url,
-            data=payment_processor.as_json(),
+            data=json_data,
             allow_redirects=False,
         )
 
         try:
             payment.transaction_id = response_dict['orderId']
 
+            if 'payMethods' in response_dict:
+                payment.set_renew_token(
+                    response_dict['payMethods']['payMethod']['value'],
+                    card_expire_year=response_dict['payMethods']['payMethod']['card']['expirationYear'],
+                    card_expire_month=response_dict['payMethods']['payMethod']['card']['expirationMonth'],
+                    card_masked_number=response_dict['payMethods']['payMethod']['card']['number'],
+                    automatic_renewal=self.recurring_payments,
+                )
+            add_extra_data(payment, {'card_response': response_dict})
+
             if response_dict['status']['statusCode'] == u'SUCCESS':
-                if 'payMethods' in response_dict:
-                    payment.set_renew_token(
-                        response_dict['payMethods']['payMethod']['value'],
-                        card_expire_year=response_dict['payMethods']['payMethod']['card']['expirationYear'],
-                        card_expire_month=response_dict['payMethods']['payMethod']['card']['expirationMonth'],
-                    )
-                payment.extra_data = json.dumps({'card_response': response_dict}, indent=2)
-                payment.save()
                 if 'redirectUri' in response_dict:
                     payment.pay_link = response_dict['redirectUri']
                     payment.save()
@@ -355,15 +403,15 @@ class PayuProvider(BasicProvider):
                         return "success"
                     return payment_processor.continueUrl
             elif response_dict['status']['statusCode'] == u'WARNING_CONTINUE_CVV':
-                payment.extra_data = json.dumps({'cvv_url': response_dict['redirectUri']}, indent=2)
-                payment.save()
+                add_extra_data(payment, {'cvv_url': response_dict['redirectUri']})
                 return payment.get_payment_url()
             elif response_dict['status']['statusCode'] == u'WARNING_CONTINUE_3DS':
+                add_extra_data(payment, {'3ds_url': response_dict['redirectUri']})
                 return response_dict['redirectUri']
         except KeyError:
             pass
 
-        raise PayuApiError(response_dict)
+        raise PayuApiError(response_dict, json_data)
 
     # Method that returns all pay methods
 
@@ -422,16 +470,34 @@ class PayuProvider(BasicProvider):
             signature = m.hexdigest()
             if incoming_signature == signature:  # and not payment.status == PaymentStatus.CONFIRMED:
                 data = json.loads(request.body.decode("utf8"))
-                status = data['order']['status']
-                status_map = {
-                    'COMPLETED': PaymentStatus.CONFIRMED,
-                    'PENDING': PaymentStatus.WAITING,
-                    'WAITING_FOR_CONFIRMATION': PaymentStatus.WAITING,
-                    'CANCELED': PaymentStatus.REJECTED,
-                    'NEW': '',
-                }
-                payment.change_status(status_map[status])
-                return HttpResponse("ok", status=200)
+                add_new_status(payment, data)
+                if 'refund' in data:
+                    refunded_price = dequantize_price(
+                        data['refund']['amount'],
+                        data['refund']['currencyCode'],
+                    )
+                    if data['refund']['status'] == 'FINALIZED' and refunded_price == payment.total:
+                        payment.message += data['refund']['reasonDescription']
+                        payment.change_status(PaymentStatus.REFUNDED)
+                        return HttpResponse("ok", status=200)
+                    else:
+                        raise Exception("Partial refund or refund status not supported", data)
+                else:
+                    status = data['order']['status']
+                    status_map = {
+                        'COMPLETED': PaymentStatus.CONFIRMED,
+                        'PENDING': PaymentStatus.INPUT,
+                        'WAITING_FOR_CONFIRMATION': PaymentStatus.INPUT,
+                        'CANCELED': PaymentStatus.REJECTED,
+                        'NEW': PaymentStatus.WAITING,
+                    }
+                    if PaymentStatus.CONFIRMED and 'totalAmount' in data['order']:
+                        payment.captured_amount = dequantize_price(
+                            data['order']['totalAmount'],
+                            data['order']['currencyCode'],
+                        )
+                    payment.change_status(status_map[status])
+                    return HttpResponse("ok", status=200)
         return HttpResponse("not ok", status=500)
 
     def process_data(self, payment, request, *args, **kwargs):
@@ -445,29 +511,33 @@ class PayuProvider(BasicProvider):
             return self.process_widget_callback(payment, renew_token, recurring="STANDARD")
         elif 'value' in request.POST:
             return self.process_widget_callback(payment, request.POST.get('value'), recurring="FIRST")
+        else:
+            return HttpResponse("request not recognized by django-payments-payu provider", status=500)
 
 
 class PaymentProcessor(object):
     "Payment processor"
 
-    def __init__(self, order, notify_url, currency, description, customer_ip, **kwargs):
+    def __init__(self, order, notify_url, currency, description, customer_ip, total, tax, **kwargs):
         self.order = order
         self.notify_url = notify_url
         self.currency = currency
         self.description = description
         self.customer_ip = customer_ip
+        self.tax = tax
         self.order_items = []
         self.external_id = None
         self.pos_id = None
-        self.total = None
+        self.total = total
 
     def get_order_items(self):
         for purchased_item in self.order:
             item = {
                 'name': purchased_item.name[:127],
                 'quantity': purchased_item.quantity,
-                'unitPrice': purchased_item.price,
+                'unitPrice': quantize_price(purchased_item.price * purchased_item.tax_rate, self.currency),
                 'currency': purchased_item.currency,
+                'subUnit': int(CURRENCY_SUB_UNIT[self.currency]),
             }
             yield item
 
@@ -485,32 +555,21 @@ class PaymentProcessor(object):
                 'phone': phone,
                 'firstName': first_name,
                 'lastName': last_name,
-                'language': lang_code
+                'language': lang_code,
             }
 
     def as_json(self):
         "Return json for the payment"
-        total = 0
-        products = []
-        order_items = self.get_order_items()
-        for i in order_items:
-            total += i['unitPrice'] * i['quantity']
-            i['subUnit'] = int(CURRENCY_SUB_UNIT[self.currency])
-            i['unitPrice'] = quantize_price(i['unitPrice'], self.currency)
-            products.append(i)
-
-        self.total = quantize_price(total, self.currency)
+        products = list(self.get_order_items())
 
         json_dict = {
             'notifyUrl': self.notify_url,
             'customerIp': self.customer_ip,
-            # 'extOrderId': self.external_id,
-            # NOTE: extOrderId prevents the payment from being submitted twice, which generates error,
-            # that is can't be easily displayed to user, so we left it blank.
+            'extOrderId': self.external_id,
             'merchantPosId': self.pos_id,
             'description': self.description,
             'currencyCode': self.currency,
-            'totalAmount': self.total,
+            'totalAmount': quantize_price(self.total, self.currency),
             'products': products,
         }
 
