@@ -87,10 +87,15 @@ class Payment(Mock):
     def get_user_email(self):
         return "foo@bar.com"
 
-    def set_renew_token(self, token, card_expire_year=None, card_expire_month=None):
+    def set_renew_token(
+            self, token, card_expire_year=None, card_expire_month=None,
+            card_masked_number=None, automatic_renewal=None,
+    ):
         self.token = token
         self.card_expire_year = card_expire_year
         self.card_expire_month = card_expire_month
+        self.card_masked_number = card_masked_number
+        self.automatic_renewal = automatic_renewal
 
 
 class TestPayuProvider(TestCase):
@@ -147,7 +152,7 @@ class TestPayuProvider(TestCase):
                 "orderId": 123,
                 "payMethods": {"payMethod": {
                     "value": 1211,
-                    "card": {"expirationYear": 2021, "expirationMonth": 1},
+                    "card": {"expirationYear": 2021, "expirationMonth": 1, "number": "1234xxx"},
                 }},
             })
             post.status_code = 200
@@ -156,6 +161,10 @@ class TestPayuProvider(TestCase):
                 self.provider.get_form(payment=self.payment)
             self.assertEqual(context.exception.args[0], 'test_redirect_uri')
             self.assertEqual(self.payment.token, 1211)
+            self.assertEqual(self.payment.card_expire_year, 2021)
+            self.assertEqual(self.payment.card_expire_month, 1)
+            self.assertEqual(self.payment.card_masked_number, "1234xxx")
+            self.assertEqual(self.payment.automatic_renewal, True)
 
     def test_redirect_payu_unknown_status(self):
         self.set_up_provider(True, False)
@@ -180,11 +189,11 @@ class TestPayuProvider(TestCase):
                     {"buyer": {
                         "email": "foo@bar.com", "language": "en", "lastName": "Bar", "firstName": "Foo", "phone": None,
                     },
-                        "description": "payment", "totalAmount": 20000, "merchantPosId": "123abc", "customerIp": "123",
+                        "description": "payment", "totalAmount": 22000, "merchantPosId": "123abc", "customerIp": "123",
                         "notifyUrl": "https://example.com/process_url/token",
-                        "products": [
-                            {"currency": "USD", "name": "foo", "quantity": 10, "unitPrice": 20, "subUnit": 100},
-                        ],
+                        "extOrderId": "bar_token",
+                        "products":
+                        [{"currency": "USD", "name": "foo", "quantity": 10, "unitPrice": 2000, "subUnit": 100}],
                         "continueUrl": "http://foo_succ.com", "currencyCode": "USD",
                     },
                 ),
@@ -213,11 +222,11 @@ class TestPayuProvider(TestCase):
                     {"buyer": {
                         "email": "foo@bar.com", "language": "en", "lastName": "Bar", "firstName": "Foo", "phone": None,
                     },
-                        "description": "payment", "totalAmount": 20000, "merchantPosId": "123abc", "customerIp": "123",
+                        "description": "payment", "totalAmount": 22000, "merchantPosId": "123abc", "customerIp": "123",
                         "notifyUrl": "https://example.com/process_url/token",
-                        "products": [
-                            {"currency": "USD", "name": "foo", "quantity": 10, "unitPrice": 20, "subUnit": 100},
-                        ],
+                        "extOrderId": "bar_token",
+                        "products":
+                        [{"currency": "USD", "name": "foo", "quantity": 10, "unitPrice": 2000, "subUnit": 100}],
                         "continueUrl": "http://foo_succ.com", "currencyCode": "USD",
                     },
                 ),
@@ -237,7 +246,11 @@ class TestPayuProvider(TestCase):
             mocked_post.return_value = post
             with self.assertRaises(PayuApiError) as context:
                 self.provider.get_form(payment=self.payment)
-            self.assertEqual(context.exception.args[0], "Unable to regain authorization token")
+            self.assertEqual(
+                context.exception.args[0],
+                "Unable to regain authorization token "
+                "{'redirectUri': 'test_redirect_uri', 'status': {'statusCode': 'UNAUTHORIZED'}, 'orderId': 123}",
+            )
 
             mocked_post.assert_called_with(
                 'http://mock.url/pl/standard/user/oauth/authorize',
@@ -275,7 +288,9 @@ class TestPayuProvider(TestCase):
         self.set_up_provider(True, True)
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_CVV"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_CVV"}, "orderId": 123}
+            )
             post.status_code = 200
             mocked_post.return_value = post
             redirect = self.provider.process_data(payment=self.payment, request=post)
@@ -289,10 +304,12 @@ class TestPayuProvider(TestCase):
                     {
                         "products":
                         [
-                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 20, "subUnit": 100}
+                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 2000, "subUnit": 100}
                         ],
+                        "extOrderId": "bar_token",
                         "buyer": {
-                            "phone": None, "email": "foo@bar.com", "lastName": "Bar", "language": "en", "firstName": "Foo",
+                            "phone": None, "email": "foo@bar.com",
+                            "lastName": "Bar", "language": "en", "firstName": "Foo",
                         },
                         "merchantPosId": "123abc", "notifyUrl": "https://example.com/process_url/token",
                         "payMethods": {
@@ -300,7 +317,8 @@ class TestPayuProvider(TestCase):
                                 "value": "bar_token", "type": "CARD_TOKEN"
                             }
                         },
-                        "totalAmount": 20000, "continueUrl": "http://foo_succ.com", "customerIp": "123", "description": "payment",
+                        "totalAmount": 22000, "continueUrl": "http://foo_succ.com",
+                        "customerIp": "123", "description": "payment",
                         "recurring": "STANDARD", "currencyCode": "USD"
                     },
                 ),
@@ -313,7 +331,9 @@ class TestPayuProvider(TestCase):
         self.payment.extra_data = json.dumps({'cvv_url': "foo_url"})
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123},
+            )
             post.status_code = 200
             mocked_post.return_value = post
             form = self.provider.get_form(payment=self.payment)
@@ -327,7 +347,9 @@ class TestPayuProvider(TestCase):
         self.set_up_provider(True, False)
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_3DS"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_3DS"}, "orderId": 123}
+            )
             post.status_code = 200
             mocked_post.return_value = post
             with self.assertRaises(RedirectNeeded) as context:
@@ -338,10 +360,17 @@ class TestPayuProvider(TestCase):
                 data=JSONEquals(
                     {
                         "merchantPosId": "123abc", "continueUrl": "http://foo_succ.com",
-                        "buyer": {"lastName": "Bar", "phone": None, "email": "foo@bar.com", "firstName": "Foo", "language": "en"},
-                        "description": "payment", "notifyUrl": "https://example.com/process_url/token", "totalAmount": 20000,
+                        "buyer": {
+                            "lastName": "Bar", "phone": None,
+                            "email": "foo@bar.com", "firstName": "Foo", "language": "en"
+                        },
+                        "description": "payment",
+                        "notifyUrl": "https://example.com/process_url/token", "totalAmount": 22000,
                         "currencyCode": "USD",
-                        "products": [{"name": "foo", "quantity": 10, "subUnit": 100, "currency": "USD", "unitPrice": 20}],
+                        "extOrderId": "bar_token",
+                        "products": [{
+                            "name": "foo", "quantity": 10, "subUnit": 100, "currency": "USD", "unitPrice": 2000,
+                        }],
                         "customerIp": "123"
                     },
                 ),
@@ -410,12 +439,33 @@ class TestPayuProvider(TestCase):
         self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
         self.assertEqual(self.payment.captured_amount, Decimal('0'))
 
+    def test_process_notification_total_amount(self):
+        """ Test processing PayU notification if it captures correct amount """
+        self.set_up_provider(True, True)
+        mocked_request = MagicMock()
+        mocked_request.body = json.dumps(
+            {"order": {"status": "COMPLETED", "totalAmount": 200, "currencyCode": "USD"}},
+        ).encode("utf8")
+        mocked_request.META = {
+            'CONTENT_TYPE': 'application/json',
+            "HTTP_OPENPAYU_SIGNATURE": "signature=01a0e768ab1f762da4b955585aa4e59e;algorithm=MD5",
+        }
+        mocked_request.status_code = 200
+        ret_val = self.provider.process_data(payment=self.payment, request=mocked_request)
+        self.assertEqual(ret_val.__class__.__name__, "HttpResponse")
+        self.assertEqual(ret_val.status_code, 200)
+        self.assertEqual(ret_val.content, b"ok")
+        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
+        self.assertEqual(self.payment.captured_amount, Decimal('2'))
+
     def test_process_notification_error(self):
         """ Test processing PayU notification with wrong signature """
         self.set_up_provider(True, True)
         mocked_request = MagicMock()
         mocked_request.body = b'{}'
-        mocked_request.META = {'CONTENT_TYPE': 'application/json', "HTTP_OPENPAYU_SIGNATURE": "signature=foo;algorithm=MD5"}
+        mocked_request.META = {
+            'CONTENT_TYPE': 'application/json', "HTTP_OPENPAYU_SIGNATURE": "signature=foo;algorithm=MD5"
+        }
         ret_val = self.provider.process_data(payment=self.payment, request=mocked_request)
         self.assertEqual(ret_val.__class__.__name__, "HttpResponse")
         self.assertEqual(ret_val.status_code, 500)
@@ -452,12 +502,18 @@ class TestPayuProvider(TestCase):
                 allow_redirects=False,
                 data=JSONEquals(
                     {
-                        "recurring": "FIRST", "customerIp": "123", "totalAmount": 20000,
+                        "recurring": "FIRST", "customerIp": "123", "totalAmount": 22000,
                         "description": "payment",
-                        "products": [{"name": "foo", "subUnit": 100, "currency": "USD", "unitPrice": 20, "quantity": 10}],
+                        "extOrderId": None,
+                        "products": [{
+                            "name": "foo", "subUnit": 100, "currency": "USD", "unitPrice": 2000, "quantity": 10
+                        }],
                         "continueUrl": "http://foo_succ.com", "merchantPosId": "123abc", "currencyCode": "USD",
                         "payMethods": {"payMethod": {"value": "renew_token", "type": "CARD_TOKEN"}},
-                        "buyer": {"firstName": "Foo", "email": "foo@bar.com", "language": "en", "phone": None, "lastName": "Bar"},
+                        "buyer": {
+                            "firstName": "Foo", "email": "foo@bar.com",
+                            "language": "en", "phone": None, "lastName": "Bar",
+                        },
                         "notifyUrl": "https://example.com/process_url/token",
                     }
                 ),
@@ -471,7 +527,9 @@ class TestPayuProvider(TestCase):
         self.set_up_provider(True, True)
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123}
+            )
             post.status_code = 200
             mocked_post.return_value = post
             redirect = self.provider.process_data(payment=self.payment, request=mocked_post)
@@ -484,10 +542,12 @@ class TestPayuProvider(TestCase):
                     {
                         "products":
                         [
-                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 20, "subUnit": 100}
+                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 2000, "subUnit": 100}
                         ],
+                        "extOrderId": "bar_token",
                         "buyer": {
-                            "phone": None, "email": "foo@bar.com", "lastName": "Bar", "language": "en", "firstName": "Foo",
+                            "phone": None, "email": "foo@bar.com",
+                            "lastName": "Bar", "language": "en", "firstName": "Foo",
                         },
                         "merchantPosId": "123abc", "notifyUrl": "https://example.com/process_url/token",
                         "payMethods": {
@@ -495,7 +555,8 @@ class TestPayuProvider(TestCase):
                                 "value": "bar_token", "type": "CARD_TOKEN"
                             }
                         },
-                        "totalAmount": 20000, "continueUrl": "http://foo_succ.com", "customerIp": "123", "description": "payment",
+                        "totalAmount": 22000, "continueUrl": "http://foo_succ.com",
+                        "customerIp": "123", "description": "payment",
                         "recurring": "STANDARD", "currencyCode": "USD"
                     },
                 ),
@@ -510,7 +571,9 @@ class TestPayuProvider(TestCase):
         self.provider.card_on_file = True
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "http://test_redirect_uri.com/", "status": {"statusCode": "SUCCESS"}, "orderId": 123}
+            )
             post.status_code = 200
             mocked_post.return_value = post
             redirect = self.provider.process_data(payment=self.payment, request=mocked_post)
@@ -523,10 +586,12 @@ class TestPayuProvider(TestCase):
                     {
                         "products":
                         [
-                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 20, "subUnit": 100}
+                            {"currency": "USD", "quantity": 10, "name": "foo", "unitPrice": 2000, "subUnit": 100}
                         ],
+                        "extOrderId": "bar_token",
                         "buyer": {
-                            "phone": None, "email": "foo@bar.com", "lastName": "Bar", "language": "en", "firstName": "Foo",
+                            "phone": None, "email": "foo@bar.com",
+                            "lastName": "Bar", "language": "en", "firstName": "Foo",
                         },
                         "merchantPosId": "123abc", "notifyUrl": "https://example.com/process_url/token",
                         "payMethods": {
@@ -534,7 +599,8 @@ class TestPayuProvider(TestCase):
                                 "value": "bar_token", "type": "CARD_TOKEN"
                             }
                         },
-                        "totalAmount": 20000, "continueUrl": "http://foo_succ.com", "customerIp": "123", "description": "payment",
+                        "totalAmount": 22000, "continueUrl": "http://foo_succ.com",
+                        "customerIp": "123", "description": "payment",
                         "cardOnFile": "STANDARD_CARDHOLDER", "currencyCode": "USD"
                     },
                 ),
@@ -561,7 +627,9 @@ class TestPayuProvider(TestCase):
         self.set_up_provider(True, True)
         with patch('requests.post') as mocked_post:
             post = MagicMock()
-            post.text = '{"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_CVV"}, "orderId": 123}'
+            post.text = json.dumps(
+                {"redirectUri": "test_redirect_uri", "status": {"statusCode": "WARNING_CONTINUE_CVV"}, "orderId": 123}
+            )
             post.status_code = 200
             mocked_post.return_value = post
             redirect = self.provider.auto_complete_recurring(self.payment)
